/*
  zBoy - A GameBoy emulator for Windows and Linux
  Copyright (C) Mateusz Viste 2010, 2011, 2012, 2013, 2014

  ----------------------------------------------------------------------
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.
  ----------------------------------------------------------------------
*/


#ifdef __zboy4linux__
  #pragma message("Compiling zBoy for Linux...")
#endif

#ifdef __zboy4windows__
  #pragma message("Compiling zBoy for Windows...")
  #define printf(...) /* Ignore any printf() calls on windows */
  /* void printf(char *s, ...) {
  } */
#endif


/* include some standard C headers */
#include <stdio.h>      /* for printf() and fprintf() */
#include <stdlib.h>     /* for atoi() and exit(), random() and randomize() */
#include <time.h>       /* for time-related functions (clock, time...) */
#include <stdint.h>     /* for uint8_t type */
#include <signal.h>

#include <pthread.h>    /* for multi-threading capabilities - not on DOS, obviously */
#include "SDL/SDL.h"

#define pVer "0.53"
#define pDate "2010,2011,2012,2013,2014"

/*  The options below make zBoy to be compiled in debug mode, and optionally trace CPU activity to a trace file */
/* #define DEBUGMODE */
/* #define DEBUGTRACEFILE */
/* #define MEMDUMP_AT_EXIT */

/* Produce a stand-alone exe file with an embedded ROM file */
/* #define EMBEDROM "embedrom/dracula.c" */
/* #define EMBEDROM "embedrom/pacman.c" */
/* #define EMBEDROM "embedrom/drmario.c" */

#ifndef DEBUGMODE   /* If not in debugmode, then disable PrintDebug */
  /* #define PrintDebug printf */
  /* #define PrintDebug(...) */
  void PrintDebug(char *s, ...) {
    s = s;  /* just to make gcc warnings shut up */
  }
  #else
  #define PrintDebug printf
#endif

#include "zboystructs.h"

struct RomInformations RomInfos;

char CartTypes[256][256];
signed char QuitEmulator;
signed int AskForScreenshot = 0;
unsigned int LastScreenshotTime = 0;
uint8_t NetPlay = 0;
char *NetPlayPeer;
char *NetPlayBindaddr = NULL;

unsigned int GraphicWidth, GraphicHeight;
char UserMessage[1024];
unsigned int UserMessageFramesLeft = 0;

/* the zboyversion() function is used by many modules to display the app version */
char *zboyversion() {
  return(pVer);
}
/* the zboyversion() function is used by many modules to display the app date */
char *zboydate() {
  return(pDate);
}

#include "about.h"
#include "declares.h"         /* Some SUB/FUNCTION declarations */
#include "binary.h"           /* Provides basic support to binary notation (eg. bx01100111) */
#include "adjtiming.c"        /* AdjustTiming() */
#include "config.h"           /* zBoy_loadconf() & zBoy_saveconf() */
#include "numconv.c"          /* UbyteToByte() and DwordVal() */
#include "presskey.c"         /* PressAnyKey() */
#include "getsavedir.h"       /* Provides the GetSaveDir() call */
#include "getsnapshotfilename.c"
#include "mmu.c"              /* Include Memory Management Unit emulation core */
#include "hiscores.c"         /* Provides SaveHiScoresMem() and LoadHiScoresMem() */
#include "cpu-ints.c"         /* Interrupts routines */
#include "video.c"            /* Video emulation core */
#include "loadpal.h"          /* LoadPalette() */
#include "joypad.c"           /* Joypad interface emulation */
#include "cpu-timer.c"        /* Include timer emulation */
#include "cpu-divider.c"      /* Include divider emulation */
#ifdef DEBUGMODE
  #include "debug.c"          /* Includes some debugging routines */
#endif
#include "cpu-z80.c"          /* Z80 CPU emulation core */
#include "battery.c"          /* Provides LoadBattRAM() and SaveBattRAM() */
#include "font.c"             /* Includes font data used by zBoy to display onscreen messages */
#include "wordwrap.h"         /* WordWrap(), required by PrintMsg() */
#include "printmsg.c"         /* PrintMsg(text AS STRING) */
#include "loadrom.h"          /* LoadROM(RomFile AS STRING) */
#include "serial.c"           /* Serial interface emulation */
#include "savepcxfile.c"      /* Screenshot routine SavePcxFile() */
#include "save.c"             /* SaveGame() and LoadGame() */
#include "reset.c"            /* ResetEmulator() */
#include "ssaver.c"           /* InhibitScreenSaver() */
#include "about.c"            /* printhelp() */


int zboy(char* roms) {
  char *NoYes[2] = {"No", "Yes"};
  struct zboyparamstype zboyparams;
  int x;
  int UsedCycles;

  /* Prepare SDL-related variables */
  SDL_Surface *screen;      /* virtual SDL screen area */
  SDL_Color palcolors[256]; /* SDL palette */

  /* Set zBoy parameters to default values */
  zboyparams.ShowRomInfos = 0;
  zboyparams.NoScreenSaver = 1;
  zboyparams.ShowFPS = 0;
  zboyparams.NoSpeedLimit = 1;
  zboyparams.HiScoresMem = 0;
  zboyparams.NoCpuIdle = 1;
  zboyparams.scalingalgo = 0;  /* By default, use Scale2x scaling... */
  zboyparams.GraphicScaleFactor = 0;          /* ...with 4x magnification. */
  zboyparams.fpslimit = 60;
  zboyparams.bootlogo = 0;
  zboyparams.palette = 0; /* Set up the default palette */
  zboyparams.colorization = 1;
  RomInfos.MemoryROM_PTR = MemoryROM;

  /* Here I am preloading a default configuration for input keys mapping */
  zboyparams.key_up = keystring2int("kb.up");
  zboyparams.key_down = keystring2int("kb.down");
  zboyparams.key_left = keystring2int("kb.left");
  zboyparams.key_right = keystring2int("kb.right");
  zboyparams.key_start = keystring2int("kb.ret");
  zboyparams.key_select = keystring2int("kb.tab");
  zboyparams.key_a = keystring2int("kb.lctrl");
  zboyparams.key_b = keystring2int("kb.lalt");
  zboyparams.key_turboa = keystring2int("none");
  zboyparams.key_turbob = keystring2int("none");
  zboyparams.key_bckg = keystring2int("kb.f1");
  zboyparams.key_sprt = keystring2int("kb.f2");
  zboyparams.key_wind = keystring2int("kb.f3");
  zboyparams.key_save = keystring2int("kb.f5");
  zboyparams.key_load = keystring2int("kb.f7");
  zboyparams.key_asht = keystring2int("kb.f8");
  zboyparams.key_shot = keystring2int("kb.f9");
  zboyparams.key_rset = keystring2int("kb.f10");
  zboyparams.key_quit = keystring2int("kb.esc");

  zBoy_loadconfig(&zboyparams); /* (try to) load parameters from the zBoy configuration file */
 
  snprintf(RomInfos.Filename, 1022, roms);

  zBoy_fixconfig(&zboyparams);  /* check the configuration, and fix it if needed */
  switch (zboyparams.palette) {
    case 1:
      LoadPalette("peasoup", ScreenPalette32); /* Load the palette */
      break;
    default:
      LoadPalette("gbmicro", ScreenPalette32); /* Load the palette */
      break;
  }
  /* Set up input keys definitions */
  if (RedefineInputKeys(int2keystring(zboyparams.key_up), int2keystring(zboyparams.key_down), int2keystring(zboyparams.key_left), int2keystring(zboyparams.key_right), int2keystring(zboyparams.key_start), int2keystring(zboyparams.key_select), int2keystring(zboyparams.key_a), int2keystring(zboyparams.key_b), int2keystring(zboyparams.key_turboa), int2keystring(zboyparams.key_turbob)) != 0) printf("error redefining input keys\n");;
  if (RedefineFunctionKeys(int2keystring(zboyparams.key_bckg), int2keystring(zboyparams.key_sprt), int2keystring(zboyparams.key_wind), int2keystring(zboyparams.key_save), int2keystring(zboyparams.key_load), int2keystring(zboyparams.key_asht), int2keystring(zboyparams.key_shot), int2keystring(zboyparams.key_rset), int2keystring(zboyparams.key_quit)) != 0) printf("error redefining function keys\n");

    GraphicWidth = 160 * zboyparams.GraphicScaleFactor;
    GraphicHeight = 144 * zboyparams.GraphicScaleFactor;

  /* Initialize the SDL library... */
  SDL_Init(SDL_INIT_VIDEO);
  SDL_ShowCursor(SDL_DISABLE); /* Disable (hide) mouse cursor */
  
  screen = SDL_SetVideoMode(GraphicWidth, GraphicHeight, 8, SDL_SWSURFACE);
  if (screen == NULL)
  {
     printf("Couldn't set video mode: %s\n", SDL_GetError());
     return(1);
  }

  SDL_WM_SetCaption("zBoy", "zBoy");

  InitScreenOldBuffer();   /* Init the screen buff used for LCD change detection (will force the complete draw of the 1st frame) */

  /* Put palette into SDL now, in case we need to display an error message */
  for (x = 0; x < 256; x++) {
    palcolors[x].r = (ScreenPalette32[x] >> 16) & 0xFF;
    palcolors[x].g = (ScreenPalette32[x] >> 8) & 0xFF;
    palcolors[x].b = ScreenPalette32[x] & 0xFF;
  }
  SDL_SetPalette(screen, SDL_LOGPAL|SDL_PHYSPAL, palcolors, 0, 255);

  #ifndef EMBEDROM
  if (RomInfos.Filename[0] == 0) 
  {
    printf("No ROM filename provided!\n");
    sprintf(UserMessage, "\n  *** ZBOY V%s ***\n\nUSAGE: ZBOY FILE.GB\n\nSEE ZBOY.TXT FOR DETAILED INSTRUCTIONS.", pVer);
    PrintMsg(UserMessage, 0);
    RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
    PressAnyKey();
    return(0);
  }

  printf("Loading the ROM data into memory (%s)...\n", RomInfos.Filename);
  x = LoadRom(&RomInfos, UserMessage);
    
  if (x != 0) 
  {
    char *errstr = LoadRomErrStr(x);
    printf("Error: Could not load the ROM file. [%d]\n", x);
    PrintMsg(errstr, 0);
    RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
    PressAnyKey();
    return(0);
  }
  #endif
  
  if (zboyparams.colorization != 0) CheckForColorizedPalette(RomInfos.CrcSum , ScreenPalette32);
  printf("Uploading palette into the virtual screen...\n");
  for (x = 0; x < 256; x++) {
    palcolors[x].r = (ScreenPalette32[x] >> 16) & 0xFF;
    palcolors[x].g = (ScreenPalette32[x] >> 8) & 0xFF;
    palcolors[x].b = ScreenPalette32[x] & 0xFF;
  }
  SDL_SetPalette(screen, SDL_LOGPAL|SDL_PHYSPAL, palcolors, 0, 255);

  switch (RomInfos.MbcModel) {  /* Check the MBC chip type, and select the right Read/Write functions */
    case 0:   /* ROM ONLY */
      MemoryWriteSpecial = &MBC0_MemoryWrite;
      MemoryReadSpecial = &MBC0_MemoryRead;
      break;
    case 1:   /* MBC1 */
      MemoryWriteSpecial = &MBC1_MemoryWrite;
      MemoryReadSpecial = &MBC1_MemoryRead;
      break;
    case 2:   /* MBC2 */
      MemoryWriteSpecial = &MBC2_MemoryWrite;
      MemoryReadSpecial = &MBC2_MemoryRead;
      break;
    case 3:   /* MBC3 */
      MemoryWriteSpecial = &MBC3_MemoryWrite;
      MemoryReadSpecial = &MBC3_MemoryRead;
      break;
    case 5:   /* MBC5 */
      MemoryWriteSpecial = &MBC5_MemoryWrite;
      MemoryReadSpecial = &MBC5_MemoryRead;
      break;
    default:  /* Any other (not supported) memory controller */
      PrintMsg("ERROR: THE MEMORY CONTROLLER USED BY THIS ROM IS NOT SUPPORTED (YET) BY ZBOY. SORRY.", 0);
      RefreshScreen(screen, 0, 159, 0, 143, &zboyparams);
      PressAnyKey();
      return(0);
      break;
  }

  /* Set up the window's title */
  if (RomInfos.NiceTitle[0] != 0) {
      snprintf(zboyparams.windowtitle, 256, "%s (zBoy)", RomInfos.NiceTitle);
    } else {
      snprintf(zboyparams.windowtitle, 256, "zBoy");
  }
  SDL_WM_SetCaption(zboyparams.windowtitle, "zBoy");

  srand(time(NULL));  /* Seed the random generator */

  /*InitRAM() */    /* Init RAM areas to random values (as in a real GameBoy) */
  if ((RomInfos.Battery != 0) && (RomInfos.RamSize > 0)) LoadBattRAM();  /* check for battery-saved RAM */

  BootRomEnabledFlag = zboyparams.bootlogo; /* Set up the boot logo flag */
  ResetEmulator(); /* Fire up the Zilog80 monster, and init all required variables */

  if (zboyparams.HiScoresMem != 0) LoadHiScoresMem();   /* Check out if need to look after hiscores */

  printf("Starting emulation...\n");

  do {
    UsedCycles = CpuExec();
    TotalCycles += UsedCycles;    /* Increment the global cycles counter */
    uTimer(UsedCycles);           /* Update uTimer */
    incDivider(UsedCycles);       /* Increment the DIV register */
    VideoSysUpdate(UsedCycles, screen, &zboyparams);   /* Update Video subsystem */

    CheckSerialLink(UsedCycles);  /* Check serial link (not supported yet) */
    CheckJoypad(UsedCycles, &zboyparams);  /* Update the Joypad register */
    CheckInterrupts();  /* Check if any interrupt has been requested since last time */
/*    #ifdef DEBUGMODE
      if ((TIMER - LastDebugOnscreen) > 0.5) { */
/*        DebugSpriteDump(1); */  /* perform some debug dumps onscreen (OBJ table at &h8000) */
/*        DebugSpriteDump(2); */  /* perform some debug dumps onscreen (OBJ table at &h8800) */
/*        DebugPaletteDump(); */  /* Dump system's palette onscreen */
/*        DebugShowCPUspeed(); */ /* Calc current CPU speed and put it onscreen */
/*        DebugShowCpuRegisters(); */ /* Display state of CPU registers */
/*        DebugOnscreen(); */     /* Display debug buffer informations onscreen */
/*        LastDebugOnscreen = TIMER;
      }
    #endif */
  } while (QuitEmulator == 0);

  if ((RomInfos.Battery != 0) && (RomInfos.RamSize > 0)) SaveBattRAM();  /* check for battery-saved RAM */
  if (zboyparams.HiScoresMem != 0) SaveHiScoresMem();

  #ifdef MEMDUMP_AT_EXIT
    DIM AS UINTEGER TempUINT             /* ----------------- */
    OPEN "memdump.bin" FOR OUTPUT AS #1  /* This is a memory  */
    FOR TempUINT = &h0000 TO &hFFFF      /* dump routine.     */
      MemoryRead(TempUINT, x)            /*     --  --        */
      PUT #1,, x                         /* Don't include in  */
    NEXT TempUINT                        /* official release! */
    CLOSE #1                             /* ----------------- */
  #endif

  #ifdef DEBUGMODE
    if (QuitEmulator != 0) {
      DebugOnScreen();
      PressAnyKey();
    }
  #endif

  fflush(stdin); /* Flush keyb buffer */
  QuitEmulator = 0;

  return(0);
}
